#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ingest/T2ComputeModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 19.11.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from pydantic import Field
from typing import Sequence, Union, Type
from ampel.model.ingest.T2IngestModel import T2IngestModel
from ampel.abstract.ingest.AbsT2Ingester import AbsT2Ingester
from ampel.model.UnitModel import UnitModel


class T2ComputeModel(UnitModel):
	# Override of 'unit' to enable alias
	#: Add :class:`T2 documents <ampel.content.T2Document.T2Document>` to the database
	unit: Union[str, Type[AbsT2Ingester]] = Field(..., alias='ingester')
	# config (t2 ingester config [from UnitModel])
	#: Specification of :class:`T2 documents <ampel.content.T2Document.T2Document>` to create
	units: Sequence[T2IngestModel]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/load/T3SimpleDataLoader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.12.2019
# Last Modified Date: 16.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Iterable, Union, Iterator
from ampel.type import StockId, StrictIterable
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.t3.load.AbsT3Loader import AbsT3Loader


class T3SimpleDataLoader(AbsT3Loader):
	"""Load all requested documents for the selected stocks"""

	def load(self,
		stock_ids: Union[StockId, Iterator[StockId], StrictIterable[StockId]]
	) -> Iterable[AmpelBuffer]:

		return self.db_content_loader.load(
			stock_ids = stock_ids,
			directives = self.directives,
			channel = self.channel
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/AmpelBuffer.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 31.05.2018
# Last Modified Date: 17.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Dict, List, Any, TypedDict, Literal
from ampel.type import StockId
from ampel.content.StockDocument import StockDocument
from ampel.content.DataPoint import DataPoint
from ampel.content.Compound import Compound
from ampel.content.T2Document import T2Document
from ampel.content.LogDocument import LogDocument

# Please update BufferKey on AmpelBuffer udpates
# There is currently unfortunately no way of extracting a Literal out of a TypedDict
BufferKey = Literal['id', 'stock', 't0', 't1', 't2', 'log', 'extra']

class AmpelBuffer(TypedDict, total=False):
	"""
	Content bundle used to build :class:`~ampel.view.SnapView.SnapView`.
	
	This is a dict containing 1 or more of the following items:
	"""
	# Could stock be of type List[StockDocument] to enable hybrid/dual transients ?
	id: StockId
	stock: Optional[StockDocument]
	t0: Optional[List[DataPoint]]
	t1: Optional[List[Compound]]
	t2: Optional[List[T2Document]]
	log: Optional[List[LogDocument]]
	extra: Optional[Dict[str, Any]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/load/AbsT3Loader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.12.2019
# Last Modified Date: 31.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, Iterable, Sequence, Optional, Dict, List, Iterator

from ampel.base import abstractmethod, AmpelBaseModel
from ampel.type import StockId, ChannelId, StrictIterable
from ampel.core.AdminUnit import AdminUnit, AmpelContext
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.db.DBContentLoader import DBContentLoader
from ampel.model.t3.LoaderDirective import LoaderDirective
from ampel.log.AmpelLogger import AmpelLogger

# Data loaders need access to the ampel db (and hence inherits AdminUnit)
class AbsT3Loader(AdminUnit, abstract=True):
	"""
	Base class for loading documents associated with a set of stocks.
	"""

	logger: AmpelLogger
	#: Specification of documents to load. If these are supplied as strings,
	#: they will be resolved by retrieving the corresponding alias from the
	#: Ampel config.
	directives: Sequence[LoaderDirective]
	#: Channels to load documents for
	channel: Optional[
		Union[
			ChannelId,
			AnyOf[ChannelId],
			AllOf[ChannelId],
			OneOf[ChannelId]
		]
	]


	def __init__(self, context: AmpelContext, **kwargs) -> None:
		# Note: 'directives' in kwargs can contain strings which will be
		# resolved by retrieving the associated alias from the ampel config
		directives: List[Dict] = []

		# Resolve directive aliases
		for el in kwargs.get('directives', []):
			if isinstance(el, str):
				d = context.config.get(f"alias.t3.%{el}", dict)
				if d: # mypy does not yet support type inference using the walrus operator
					directives.append(d)
				else:
					raise ValueError(f"LoaderDirective alias '{el}' not found in ampel config")
			else:
				directives.append(el)

		kwargs['directives'] = tuple(directives)

		# No need to save context as instance variable
		AmpelBaseModel.__init__(self, **kwargs)

		self.db_content_loader = DBContentLoader(
			context=context, logger=self.logger
		)


	@abstractmethod
	def load(self,
		stock_ids: Union[StockId, Iterator[StockId], StrictIterable[StockId]]
	) -> Iterable[AmpelBuffer]:
		"Load documents for the selected stocks"
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/StockT2Ingester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 23.03.2020
# Last Modified Date: 11.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from pymongo import UpdateOne
from typing import Union, List, Tuple
from ampel.type import StockId, ChannelId
from ampel.enum.T2SysRunState import T2SysRunState
from ampel.abstract.ingest.AbsT2Ingester import AbsT2Ingester
from ampel.abstract.ingest.AbsStockT2Ingester import AbsStockT2Ingester
from ampel.abstract.ingest.AbsStockT2Compiler import AbsStockT2Compiler
from ampel.ingest.compile.StockT2Compiler import StockT2Compiler


class StockT2Ingester(AbsStockT2Ingester):

	# override
	compiler: AbsStockT2Compiler = StockT2Compiler()

	def ingest(self,
		stock_id: StockId,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]]
	) -> None:

		optimized_t2s = self.compiler.compile(chan_selection)
		now = int(time())

		# Loop over t2 units to be created
		for (t2_unit_id, run_config), chans in optimized_t2s.items():

			jchan, chan_add_to_set = AbsT2Ingester.build_query_parts(chans)

			# Append update operation to bulk list
			self.updates_buffer.add_t2_update(
				UpdateOne(
					# Matching search criteria
					{
						'stock': stock_id,
						'unit': t2_unit_id,
						'config': run_config,
						'col': 'stock',
						'link': stock_id,
					},
					{
						# Attributes set if no previous doc exists
						'$setOnInsert': {
							'stock': stock_id,
							'tag': self.tags,
							'unit': t2_unit_id,
							'config': run_config,
							'status': T2SysRunState.NEW.value,
						},
						# Journal and channel update
						'$addToSet': {
							'channel': chan_add_to_set,
							'journal': {
								'tier': self.tier,
								'dt': now,
								'channel': jchan
							}
						}
					},
					upsert=True
				)
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File				: Ampel-core/ampel/t3/select/T3FilteringStockSelector.py
# License			: BSD-3-Clause
# Author			: Jakob van Santen <jakob.van.santen@desy.de>
# Date				: 02.08.2020
# Last Modified Date: 02.08.2020
# Last Modified By	: Jakob van Santen <jakob.van.santen@desy.de>

from itertools import islice
from typing import Sequence, Dict, List, Any, Union, Optional, Generator

from ampel.type import StockId
from ampel.t3.select.T3StockSelector import T3StockSelector
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.t3.T2FilterModel import T2FilterModel
from ampel.db.query.general import build_general_query
from ampel.db.query.utils import match_array


class T3FilteringStockSelector(T3StockSelector):
	"""
	Selector subclass that filters stocks based on their latest T2 results.
	
	Example configuration::
		
		{
			...
			"t2_filter": [
				{
					"unit": "T2SNCosmo",
					"match": {
						"fit_acceptable": True,
						"fit_results.c": {"$gt": 1},
					},
				}
			]
		}
	"""

	t2_filter: Union[T2FilterModel, AllOf[T2FilterModel], AnyOf[T2FilterModel]]
	chunk_size: int = 200

	# Override/Implement
	def fetch(self) -> Generator[Dict[str,Any], None, None]:

		# Execute query on T0 collection to get target stocks
		if not (cursor := super().fetch()):
			return None

		# Execute aggregation on T2 collection to get matching subset of stocks
		# NB: aggregate in chunks to avoid the 100 MB aggregation memory limit
		input_count = 0
		output_count = 0
		while (stock_ids := [doc['_id'] for doc in islice(cursor, self.chunk_size)]):
			count = 0
			for count, doc in enumerate(
				self.context.db.get_collection('t2').aggregate(
					self._t2_filter_pipeline(stock_ids),
				),
				1
			):
				yield doc

			input_count += len(stock_ids)
			output_count += count

		self.logger.info(f"{output_count}/{input_count} stocks passed filter criteria")



	def _build_match(self, f: Union[T2FilterModel, AllOf[T2FilterModel], AnyOf[T2FilterModel]]) -> Dict[str,Any]:
		if isinstance(f, T2FilterModel):
			return {f"{f.unit}.{k}": v for k, v in f.match.items()}
		elif isinstance(f, AllOf):
			return {'$and': [self._build_match(el) for el in f.all_of]}
		elif isinstance(f, AnyOf):
			return {'$or': [self._build_match(el) for el in f.any_of]}
		else:
			raise TypeError()


	def _t2_filter_pipeline(self, stock_ids: List[StockId]) -> List[Dict]:
		return self._t2_merge_pipeline(stock_ids) + [
			{'$match': self._build_match(self.t2_filter)},
			{'$project': {'_id': 1}}
		]


	def _t2_merge_pipeline(self, stock_ids: List[StockId]) -> List[Dict[str, Any]]:
		"""
		Create a pipeline for the T2 collection that yields docs whose _id is
		the stock id and whose remaining fields are the latest result for each
		T2. For example, given the following T2 docs::
			
			{
				'_id': 1,
				...
				'stock': 42,
				'unit': 'Unit1',
				'status': 0,
				'body': [{
					...
					'results': {'thing1': 3}
				}]
			},
			{
				'_id': 2,
				...
				'stock': 42,
				'unit': 'Unit2',
				'status': 0,
				'body': [{
					...
					'results': {'thing2': 7}
				}]
			}
		
		the pipeline yields a single doc of the form::
			
			{
				'_id': 42,
				'Unit1': {'thing1': 3},
				'Unit2': {'thing2': 7}
			}
		"""
		# NB: we reuse the general query here to ensure that we only process
		# T2s associated with the requested channels
		match = {
			'status': 0,
			'unit': match_array(_all_units(self.t2_filter)),
			**build_general_query(stock_ids, self.channel, self.tag)
		}

		return [
			# select t2 docs for target stocks
			{
				'$match': match
			},
			# find "latest" result for each stock and unit, where "latest" is
			# the document with the largest _id (creation time)
			{
				'$sort': {
					'stock': 1,
					'unit': 1,
					'_id': 1
				}
			},
			{'$unwind': '$body'},
			{
				'$group': {
					'_id': {
						'stock': '$stock',
						'unit': '$unit'
					},
					'result': {'$last': '$body.result'}
				}
			},
			# nest result under key named for unit, e.g.
			# {'unit': 'T2Unit', 'result': {'foo': 1}} -> {'T2Unit': {'foo': 1}}
			{
				'$replaceRoot': {
					'newRoot': {
						'_id': '$_id',
						'result': {
							'$arrayToObject': [
								[{'k': '$_id.unit', 'v': '$result'}]
							]
						}
					}
				}
			},
			{
				'$group': {
					'_id': '$_id.stock',
					'result': {'$mergeObjects': '$result'}
				}
			},
			# flatten document
			{'$set': {'result._id': '$_id'}},
			{
				'$replaceRoot': {
					'newRoot': '$result'
				}
			}
		]

def _all_units(filters: Union[T2FilterModel, AllOf[T2FilterModel], AnyOf[T2FilterModel]]) -> List[str]:
	"""
	Get the set of all units involved in the selection
	"""
	if isinstance(filters, T2FilterModel):
		return [filters.unit]
	elif isinstance(filters, AllOf):
		return list(set(f.unit for f in filters.all_of))
	elif isinstance(filters, AnyOf):
		# NB: AnyOf may contain AllOf
		return list(set(sum((_all_units(f) for f in filters.any_of), [])))
	else:
		raise TypeError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/logging/handlers/DefaultRecordBufferingHandler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 05.05.2020
# Last Modified Date: 05.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Dict
from ampel.log.handlers.RecordBufferingHandler import RecordBufferingHandler
from ampel.protocol.LoggingHandlerProtocol import LoggingHandlerProtocol
from ampel.type import StockId, ChannelId


class DefaultRecordBufferingHandler(RecordBufferingHandler):
	"""
	MemoryHandler-like class that can grown infinitely and features a convenience method called 'forward'
	"""

	def forward(self,
		target: LoggingHandlerProtocol,
		channel: Optional[ChannelId] = None,
		stock: Optional[StockId] = None,
		extra: Optional[Dict] = None,
		clear: bool = True
	) -> None:
		"""
		Forwards saved log records to provided logger/handler instance.
		Clears the internal record buffer.
		"""
		for rec in self.buffer:

			if rec.levelno < target.level:
				continue

			if channel:
				rec.channel = channel # type: ignore[union-attr]

			if stock:
				rec.stock = stock # type: ignore[union-attr]

			if extra:
				if 'extra' in rec.__dict__:
					rec.extra = {**rec.extra, **extra} # type: ignore
				else:
					rec.extra = extra # type: ignore

			target.handle(rec) # type: ignore

		if clear:
			self.buffer.clear()
			self.has_error = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/compile/StockT2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2020
# Last Modified Date: 01.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Set, List, Union, Tuple, Optional
from ampel.type import ChannelId
from ampel.abstract.ingest.AbsStockT2Compiler import AbsStockT2Compiler


class StockT2Compiler(AbsStockT2Compiler):
	"""
	Helper class capabable of generating a nested dict that is used as basis to create T2 documents.
	The generated structure is optimized: multiple T2 documents associated with the same stock record
	accross different channels are merged into one single T2 document
	that references all corrsponding channels.
	"""

	def compile(self,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]]
	) -> Dict[Tuple[str, Optional[int]], Set[ChannelId]]:
		"""
		TLDR: This function computes and returns a dict structure helpful for creating T2 docs.
		This computation is required since:
		* A given alert can be accepted by one filter and be rejected by the other
		* T0 filters can return a customized set of T2 units to be run (group id)
		:param chan_selection: example: {"CHAN_SN": True, "CHAN_GRB": 1, "CHAN_BH": None, "CHAN_A": -2}

		:returns: Dict[(unit, config): set(<channel ids>)]
		"""

		t2s_eff: Dict[Tuple[str, Optional[int]], Set[ChannelId]] = {}

		for chan, ingest_model in self.get_ingest_models(chan_selection):

			k = (ingest_model.unit_id, ingest_model.config)
			if k not in t2s_eff:
				t2s_eff[k] = {chan}
			else:
				t2s_eff[k].add(chan)

		return t2s_eff



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/run/project/AbsT3Projector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 21.01.2020
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Sequence
from ampel.base import AmpelBaseModel, AmpelABC, abstractmethod
from ampel.core.AmpelBuffer import AmpelBuffer


class AbsT3Projector(AmpelABC, AmpelBaseModel, abstract=True):


	@abstractmethod
	def project(self, seq: Sequence[AmpelBuffer]) -> Sequence[AmpelBuffer]:
		pass



from ampel.t3.load.T3SimpleDataLoader import T3SimpleDataLoader
from ampel.core.AmpelContext import AmpelContext
from ampel.config.AmpelConfig import AmpelConfig


def test_instantiate(core_config, patch_mongo, ampel_logger):
    """
    AbsT3Loader understands all the aliases in the ampel-core config
    """
    ctx = AmpelContext.new(AmpelConfig(core_config))
    aliases = ctx.config.get("alias.t3", dict)
    assert len(
        directives := T3SimpleDataLoader(
            ctx, logger=ampel_logger, directives=[k[1:] for k in aliases.keys()]
        ).directives
    ) == len(aliases)
    for d, value in zip(directives, aliases.values()):
        assert d.dict() == value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/db/AmpelDB.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.06.2018
# Last Modified Date: 31.01.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import secrets
from collections import defaultdict  # type: ignore[attr-defined]
from typing import Sequence, Dict, List, Any, Union, Optional, TYPE_CHECKING

from pymongo import MongoClient
from pymongo.database import Database
from pymongo.collection import Collection
from pymongo.errors import ConfigurationError

from ampel.type import ChannelId
from ampel.config.AmpelConfig import AmpelConfig
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.db.AmpelColModel import AmpelColModel
from ampel.model.db.AmpelDBModel import AmpelDBModel
from ampel.model.db.IndexModel import IndexModel
from ampel.model.db.ShortIndexModel import ShortIndexModel
from ampel.model.db.MongoClientOptionsModel import MongoClientOptionsModel
from ampel.model.db.MongoClientRoleModel import MongoClientRoleModel

if TYPE_CHECKING:
	from ampel.log.AmpelLogger import AmpelLogger

from ampel.abstract.AbsSecretProvider import AbsSecretProvider


intcol = {'t0': 0, 't1': 1, 't2': 2, 'stock': 3}

class AmpelDB(AmpelBaseModel):
	"""
	Ampel stores all information in a dedicated DB.
	This class allows to create or retrieve the underlying database collections.
	"""

	prefix: str = 'Ampel'
	databases: List[AmpelDBModel]
	mongo_uri: str
	mongo_options: MongoClientOptionsModel = MongoClientOptionsModel()
	secrets: Optional[AbsSecretProvider]

	@staticmethod
	def new(config: AmpelConfig, secrets: Optional[AbsSecretProvider] = None) -> 'AmpelDB':
		""" :raises: ValueError in case a required config entry is missing """
		return AmpelDB(
			mongo_uri = config.get('resource.mongo', str, raise_exc=True),
			secrets = secrets,
			**config.get('db', dict, raise_exc=True)
		)


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs) # type: ignore[call-arg]

		self.col_config: Dict[str, AmpelColModel] = {
			col.name: col
			for db_config in self.databases
				for col in db_config.collections
		}

		self.mongo_collections: Dict[str, Collection] = {}
		# map role with client
		self.mongo_clients: Dict[str, MongoClient] = {}


	def enable_rejected_collections(self, channel_names: Sequence[ChannelId]) -> None:
		"""
		Makes rejected collections (DB: Ampel_rej, collection: <channel_name>)
		available through standard method call AmpelDB.get_collection(channel_name)
		:param channel_names: list of channel names
		"""
		db_config = AmpelDBModel(
			name = 'rej',
			collections = [
				AmpelColModel(
					name=chan_name if isinstance(chan_name, str) else str(chan_name)
				) for chan_name in channel_names
			],
			role = MongoClientRoleModel(r='logger', w='logger')
		)

		self.databases.append(db_config)

		for col in db_config.collections:
			self.col_config[col.name] = col


	def get_collection(self, col_name: Union[int, str], mode: str = 'w') -> Collection:
		"""
		:param mode: required permission level, either 'r' for read-only or 'rw' for read-write
		If a collection does not exist, it will be created and the proper mongoDB indexes will be set.
		"""

		if isinstance(col_name, int):
			col_name = str(col_name)

		if col_name in self.mongo_collections:
			if mode in self.mongo_collections[col_name]:
				return self.mongo_collections[col_name][mode]
		else:
			if col_name not in self.col_config:
				raise ValueError(f"Unknown collection: '{col_name}'")
			self.mongo_collections[col_name] = {}

		db_config = self._get_db_config(col_name)
		role = db_config.role.dict()[mode]

		db = self._get_mongo_db(
			role=role, db_name=f"{self.prefix}_{db_config.name}"
		)

		if 'w' in mode and col_name not in db.list_collection_names():
			self.mongo_collections[col_name][mode] = self.create_collection(
				role, db.name, self.col_config[col_name]
			)
		else:
			self.mongo_collections[col_name][mode] = db.get_collection(col_name)

		return self.mongo_collections[col_name][mode]


	def _get_mongo_db(self, *, role: str, db_name: str) -> Database:
		if role not in self.mongo_clients:
			key = f'mongo/{role}'
			kwargs = {
				**(self.secrets.get(key, dict).get() if self.secrets else {}),
				**self.mongo_options.dict()
			}
			try:
				self.mongo_clients[role] = MongoClient(self.mongo_uri, **kwargs)
			except ConfigurationError as exc:
				# hint at error source
				raise ConfigurationError(exc.args[0] + f' (from secret {key})', *exc.args[1:])

		return self.mongo_clients[role].get_database(db_name)


	def _get_db_config(self, col_name: str) -> AmpelDBModel:
		return next(
			filter(
				lambda x: self.col_config[col_name] in x.collections,
				self.databases
			)
		)


	def init_db(self) -> None:

		for db_config in self.databases:
			for col_config in db_config.collections:
				self.get_collection(col_config.name)


	def create_collection(self,
		role: str, db_name: str,
		col_config: AmpelColModel, logger: Optional['AmpelLogger'] = None
	) -> Collection:
		"""
		:param resource_name: name of the AmpelConfig resource (resource.mongo) to be fed to MongoClient()
		:param db_name: name of the database to be used/created
		:param col_name: name of the collection to be created
		"""

		if logger is None:
			# Avoid cyclic import error
			from ampel.log.AmpelLogger import AmpelLogger
			logger = AmpelLogger.get_logger()
			logger.info(f"Creating {db_name} -> {col_config.name}")

		db = self._get_mongo_db(role=role, db_name=db_name)

		# Create collection with custom args
		if col_config.args:
			col = db.create_collection(
				col_config.name, **col_config.args
			)
		else:
			col = db.create_collection(col_config.name)

		if col_config.indexes:

			for idx in col_config.indexes:

				try:

					idx_params = idx.dict(exclude_unset=True)
					logger.info(f"  Creating index: {idx_params}")

					if idx_params.get('args'):
						col.create_index(
							idx_params['index'], **idx_params['args']
						)
					else:
						col.create_index(
							idx_params['index']
						)

				except Exception as e:
					logger.error(
						f"Index creation failed for '{col_config.name}' "
						f"(db: '{db_name}', args: {idx_params})",
						exc_info=e
					)

		return col


	def set_col_index(self,
		role: str, db_name: str, col_config: AmpelColModel,
		force_overwrite: bool = False, logger: Optional['AmpelLogger'] = None
	) -> None:
		"""
		:param force_overwrite: delete index if it already exists.
		This can be useful if you want to change index options (for example: sparse=True/False)
		"""

		if not logger:
			# Avoid cyclic import error
			from ampel.log.AmpelLogger import AmpelLogger
			logger = AmpelLogger.get_logger()

		if not col_config.indexes:
			logger.info(f"No index data configured for collection {col_config.name}")
			return

		db = self._get_mongo_db(role=role, db_name=db_name)

		if col_config.name not in db.list_collection_names():
			self.create_collection(role, db_name, col_config)
			return

		col = self.get_collection(col_config.name)
		col_index_info = col.index_information()
		flat_indexes = []

		for idx in col_config.indexes:

			idx_id = idx.get_id()
			flat_indexes.append(idx_id)

			if idx_id in col_index_info:
				if force_overwrite:
					logger.info(f"  Deleting existing index: {idx_id}")
					col.drop_index(idx_id)
				else:
					logger.info(f"  Skipping already existing index: {idx_id}")
					continue

			self._create_index(col, idx, logger)

		for k in col_index_info:
			if k not in flat_indexes and k != "_id_":
				logger.info(f"  Removing index {k}")
				col.drop_index(k)


	def __repr__(self) -> str:
		return "<AmpelDB>"


	@staticmethod
	def _create_index(
		col: Collection,
		index_data: Union[IndexModel, ShortIndexModel],
		logger: 'AmpelLogger'
	) -> None:

		try:

			idx_params = index_data.dict(exclude_unset=True)
			logger.info(f"  Creating index: {idx_params}")

			if idx_params.get('args'):
				col.create_index(idx_params['index'], **idx_params['args'])
			else:
				col.create_index(idx_params['index'])

		except Exception as e:
			logger.error(
				f"Index creation failed for '{col.name}' (args: {idx_params})",
				exc_info=e
			)


	def drop_all_databases(self):
		for db in self.databases:
			self._get_mongo_db(role=db.role.w, db_name=db.name).client.drop_database(f"{self.prefix}_{db.name}")
		self.mongo_collections.clear()
		self.mongo_clients.clear()


def provision_accounts(ampel_db: AmpelDB, auth: Dict[str, str] = {}) -> Dict[str, Any]:
	"""Create accounts required by the given Ampel configuration."""
	roles = defaultdict(list)
	for db in ampel_db.databases:
		name = f"{ampel_db.prefix}_{db.name}"
		roles[db.role.r].append({"db": name, "role": "read"})
		roles[db.role.w].append({"db": name, "role": "readWrite"})
	users = dict()
	admin = MongoClient(ampel_db.mongo_uri, **auth).get_database("admin")
	tag = secrets.token_hex(8)
	for name, all_roles in roles.items():
		username = f"{name}-{tag}"
		password = secrets.token_hex()
		admin.command("createUser", username, pwd=password, roles=all_roles)
		users[f"mongo/{name}"] = {"username": username, "password": password}
	return users


def revoke_accounts(ampel_db: AmpelDB, auth: Dict[str, str] = {}) -> None:
	"""Delete accounts previously created with "provision"."""
	if ampel_db.secrets is None:
		raise ValueError("No secrets configured")
	admin = MongoClient(ampel_db.mongo_uri, **auth).get_database("admin")
	roles = {role for db in ampel_db.databases for role in db.role.dict().values()}
	for role in roles:
		username = ampel_db.secrets.get(f"mongo/{role}", dict).get()["username"]
		admin.command("dropUser", username)


def list_accounts(ampel_db: AmpelDB, auth: Dict[str, str] = {}) -> Dict[str, Any]:
	"""List configured accounts and roles."""
	admin = MongoClient(ampel_db.mongo_uri, **auth).get_database("admin")
	return admin.command("usersInfo")


def init_db(ampel_db: AmpelDB, auth: Dict[str, str] = {}) -> None:
	"""Initialize Ampel databases and collections."""
	if auth:
		print("="*40)
		print("DANGEROUS THINGS ARE ABOUT TO HAPPEN!")
		print("="*40)
		if input(f"Do you really want to reinitialize databases {[ampel_db.prefix+'_'+db.name for db in ampel_db.databases]} on {ampel_db.mongo_uri}? All data will be lost. Type 'yessir' to proceed: ") == "yessir":
			mc = MongoClient(ampel_db.mongo_uri, **auth)
			for db in ampel_db.databases:
				name = f"{ampel_db.prefix}_{db.name}"
				mc.drop_database(name)
			ampel_db.init_db()
		else:
			print("cancelled")
	else:
		ampel_db.init_db()


def main() -> None:
	import os
	import sys
	from argparse import ArgumentParser
	from getpass import getpass

	import yaml

	from ampel.core import AmpelContext
	from ampel.dev.DictSecretProvider import DictSecretProvider

	parser = ArgumentParser(description="Manage access to Ampel databases")

	def from_env(key) -> Optional[str]:
		if fname := os.environ.get(key + '_FILE', None):
			with open(fname, "r") as f:
				return f.read().rstrip()
		else:
			return os.environ.get(key, None)

	common = ArgumentParser(add_help=False)
	common.add_argument(
		'config_file_path',
		help="Path to an Ampel configuration file"
	)
	common.add_argument(
		'-u',
		'--username',
		type=str,
		default=from_env('MONGO_ROOT_USERNAME'),
		help="Mongo admin username",
	)
	common.add_argument(
		'-p',
		'--password',
		type=str,
		default=from_env('MONGO_ROOT_PASSWORD'),
		help="Mongo admin password",
	)

	subparsers = parser.add_subparsers(dest="command")
	subparsers.required = True

	def add_command(func, name=None):
		p = subparsers.add_parser(
			func.__name__ if not name else name,
			help=func.__doc__,
			parents=[common],
		)
		p.set_defaults(command=func)
		return p

	add_command(provision_accounts, "provision")

	p = add_command(revoke_accounts, "revoke")
	p.add_argument(
		'secrets',
		type=DictSecretProvider.load,
		default=None,
		help="Path to yaml file containing the accounts entries to be revoked",
	)

	add_command(list_accounts, "list")

	p = add_command(init_db, "init")
	p.add_argument(
		'--secrets',
		type=DictSecretProvider.load,
		default=None,
		help="Path to yaml file with auth secrets",
	)

	args = parser.parse_args()
	if args.username is not None:
		if args.password is None:
			args.password = getpass(prompt=f"Password for Mongo user {args.username}: ")
		auth = {
			"username": args.username,
			"password": args.password,
		}
	else:
		auth = {}

	ctx = AmpelContext.load(args.config_file_path, secrets=getattr(args, "secrets", None))

	if (ret := args.command(ctx.db, auth)):
		yaml.dump(ret, sys.stdout, sort_keys=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/db/MongoClientOptionsModel.py
# License           : BSD-3-Clause
# Author            : Jakob van Santen <jakob.van.santen@desy.de>
# Date              : 08.11.2020
# Last Modified Date: 08.11.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from typing import Optional
from ampel.model.StrictModel import StrictModel

class MongoClientOptionsModel(StrictModel):
	socketTimeoutMS: Optional[int]
	connectTimeoutMS: Optional[int]
	serverSelectionTimeoutMS: Optional[int]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/dev/DictSecretProvider.py
# License           : BSD-3-Clause
# Author            : Jakob van Santen <jakob.van.santen@desy.de>
# Date              : 14.08.2020
# Last Modified Date: 14.08.2020
# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>

from typing import Any, Dict, Type, Tuple, get_args, get_origin, overload
import subprocess
import yaml

from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.model.Secret import Secret, T


class SecretWrapper(Secret[T]):

    value: T

    def __repr_args__(self):
        return {**dict(super().__repr_args__()), **{'value': '********'}}.items()

    def get(self) -> T:
        return self.value


class DictSecretProvider(AbsSecretProvider):

    @classmethod
    def load(cls, path: str) -> 'DictSecretProvider':
        """
        Load from a YAML file. If the file was encrypted with sops_, it will
        be decrypted with ``sops -d``.

        .. _sops: https://github.com/mozilla/sops
        """
        with open(path) as f:
            payload = yaml.safe_load(f)
        if "sops" in payload:
            try:
                payload = yaml.safe_load(subprocess.check_output(['sops', '-d', path]))
            except Exception as exc:
                raise RuntimeError(f"Can't read sops-encrypted file {path}") from exc
        return cls(payload)

    def __init__(self, items: Dict[str, Any]) -> None:
        """
        :param items:
        """
        self.store: Dict[str, Any] = dict(items)

    def get(self, key: str, value_type: Type[T]) -> SecretWrapper[T]:
        try:
            value = self.store[key]
        except KeyError:
            raise KeyError(f"Unknown secret '{key}'")
        if origin := get_origin(value_type):
            value_type = origin
        if not isinstance(value, value_type): # type: ignore[arg-type]
            raise ValueError(
                f"Retrieved value has not the expected type.\n"
                f"Expected: {value_type}\n"
                f"Found: {type(value)}"
            )
        return SecretWrapper(key=key, value=value)


class PotemkinSecretProvider(AbsSecretProvider):

    def get(self, key: str, value_type: Type[T]) -> SecretWrapper[T]:
        if get_origin(value_type) is tuple:
            value = tuple(t() for t in get_args(value_type))
        else:
            value = value_type() # type: ignore[assignment]
        return SecretWrapper(key=key, value=value) # type: ignore[arg-type]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/UnitConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 03.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import re, importlib
from typing import List, Union, Dict, Optional, Any
from ampel.util.collections import ampel_iter
from ampel.util.crypto import b2_short_hash
from ampel.model.StrictModel import StrictModel
from ampel.config.collector.AbsListConfigCollector import AbsListConfigCollector
from ampel.log import VERBOSE


class RemoteUnitDefinition(StrictModel):
	class_name: str
	base: List[str]


class UnitConfigCollector(AbsListConfigCollector):

	def add(self,
		arg: List[Union[Dict[str, str], str]],
		file_name: Optional[str] = None,
		dist_name: Optional[str] = None
	) -> None:

		# tolerate list containing only 1 element defined as dict
		for el in ampel_iter(arg):

			try:

				if isinstance(el, str):
					class_name = self.get_class_name(el)
					entry: Dict[str, Any] = {
						'fqn': el,
						'base': self.get_mro(el, class_name),
						'distrib': dist_name,
						'file': file_name
					}

				elif isinstance(el, dict):
					try:
						d = RemoteUnitDefinition(**el)
					except Exception:
						self.error(
							'Unsupported unit definition (dict)' +
							self.distrib_hint(file_name, dist_name)
						)
						continue

					class_name = d.class_name
					entry = {
						'base': d.base,
						'distrib': dist_name,
						'file': file_name
					}

				else:
					self.error(
						'Unsupported unit config format' +
						self.distrib_hint(file_name, dist_name)
					)
					continue

				if self.get(class_name):
					self.duplicated_entry(
						conf_key = class_name,
						section_detail = f'{self.tier} {self.conf_section}',
						new_file = file_name,
						new_dist = dist_name,
						prev_file = self.get(class_name).get('conf', 'unknown') # type: ignore
					)
					continue

				if "AmpelBaseModel" not in entry['base']:
					self.logger.info(
						f'Unrecognized base class for {self.conf_section} {class_name} ' +
						self.distrib_hint(file_name, dist_name)
					)
					return

				if self.conf_section == "admin unit":
					if "AdminUnit" not in entry['base']:
						raise ValueError(f"AdminUnit missing for admin unit {entry}")
					entry['base'].remove("AdminUnit")

				elif self.conf_section == "base unit":

					# Hash class name of T2 units
					for el in entry['base']:
						if 'T2Unit' in el:
							entry['hash'] = b2_short_hash(class_name)
							break

					if "DataUnit" not in entry['base']:
						raise ValueError(f"DataUnit missing for base unit {entry}")
					entry['base'].remove("DataUnit")

				entry['base'] = entry['base'][:-2]
				if self.verbose:
					self.logger.log(VERBOSE,
						f'Adding {self.conf_section}: {class_name}'
					)

				self.__setitem__(class_name, entry)

			except Exception as e:

				if 'class_name' not in locals():
					class_name = 'Unknown'

				self.error(
					f'Error occured while loading {self.conf_section} {class_name} ' +
					self.distrib_hint(file_name, dist_name),
					exc_info=e
				)


	@staticmethod
	def get_class_name(fqn: str) -> str:
		"""
		Returns class name for a given fully qualified name
		Note: we here assume the ampel convention that module name equals class name
		(i.e that we have one class per module)
		"""
		# pylint: disable=anomalous-backslash-in-string
		return re.sub(r'.*\.', '', fqn) # noqa


	@staticmethod
	def get_mro(module_fqn: str, class_name: str) -> List[str]:
		"""
		:param module_fqn: fully qualified name of module
		:param class_name: declared class name in the module specified by "module_fqn"
		:returns: the method resolution order (except for the first and last members
		that is of no use for our purpose) of the specified class
		"""
		UnitClass = getattr(
			# import using fully qualified name
			importlib.import_module(module_fqn),
			class_name
		)

		return [
			UnitConfigCollector.get_class_name(el.__name__)
			for el in UnitClass.__mro__[1:-1]
			if 'ampel' in el.__module__
		]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/demo/unit/base/DemoT3Unit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.06.2020
# Last Modified Date: 14.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Tuple
from ampel.type import T3AddResult
from ampel.view.SnapView import SnapView
from ampel.abstract.AbsT3Unit import AbsT3Unit
from ampel.struct.JournalTweak import JournalTweak


class DemoT3Unit(AbsT3Unit):


	def add(self, views: Tuple[SnapView, ...]) -> T3AddResult:
		for v in views:
			print(v.serialize())
			print("-" * 30)
		return JournalTweak(tag="DemoTag")


	def done(self) -> None:
		print("done")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsIngester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 21.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Optional, Any
from ampel.log.LogsBufferDict import LogsBufferDict
from ampel.db.DBUpdatesBuffer import DBUpdatesBuffer
from ampel.core.AdminUnit import AdminUnit


class AbsIngester(AdminUnit, abstract=True):

	updates_buffer: DBUpdatesBuffer
	logd: LogsBufferDict
	run_id: int
	hash: Optional[int]

	def get_stats(self, reset: bool = True) -> Optional[Dict[str, Any]]:
		return None

	def __hash__(self) -> int:
		if self.hash is None:
			return id(self)
		else:
			return self.hash

	def __eq__(self, other) -> bool:
		if isinstance(other, int):
			return self.hash == other
		if isinstance(other, AbsIngester):
			return self.hash == other.hash
		return False



import pytest

from pydantic import create_model, ValidationError

from ampel.model.StrictModel import StrictModel
from ampel.model.Secret import Secret
from ampel.core.UnitLoader import UnitLoader
from ampel.dev.DictSecretProvider import DictSecretProvider
from ampel.model.UnitModel import UnitModel


@pytest.fixture
def secrets():
    return DictSecretProvider(
        {
            "dict": {"a": 1, "b": "foo"},
            "str": "blahblablah",
            "tuple": (1, 1),
        }
    )


def resolve_secrets_args(expected_type, key):
    fields = {"seekrit": (Secret[expected_type], {"key": key})}
    model = create_model("testy", __config__=StrictModel.__config__, **fields)
    return model, model.__annotations__, model.__field_defaults__, {}


@pytest.mark.parametrize(
    "expected_type,key", [(dict, "dict"), (str, "str"), (tuple, "tuple")]
)
def test_resolve_secrets_correct_type(secrets, expected_type, key):
    assert (
        UnitLoader.resolve_secrets(secrets, *resolve_secrets_args(expected_type, key))[
            "seekrit"
        ].get()
        == secrets.store[key]
    )


@pytest.mark.parametrize(
    "expected_type,key", [(dict, "str"), (str, "tuple"), (tuple, "dict")]
)
def test_resolve_secrets_wrong_type(secrets, expected_type, key):
    with pytest.raises(ValueError):
        UnitLoader.resolve_secrets(secrets, *resolve_secrets_args(expected_type, key))


def test_validator_patching():
    """
    Model validation can be monkeypatched in a context manager (tripwire for
    changes in pydantic internals)
    """
    from functools import partial
    from contextlib import contextmanager
    from ampel.model.StrictModel import StrictModel

    class Model(StrictModel):
        name: str

    # add extra argument to be bound with partial()
    def validate(cls, values, other_arg):
        assert other_arg == "pass"
        return values

    @contextmanager
    def add_root_validator(model_class, func):
        extra_validator = (False, func)
        model_class.__post_root_validators__.append(extra_validator)
        try:
            yield
        finally:
            model_class.__post_root_validators__.remove(extra_validator)

    # outside context, extra validator is not run
    Model(name="fred")

    # inside context, extra validator runs (and raises an exception)
    with pytest.raises(ValidationError):
        with add_root_validator(Model, partial(validate, other_arg="fail")):
            Model(name="fred")
    # or passes, if configured to do so
    with add_root_validator(Model, partial(validate, other_arg="pass")):
        Model(name="fred")

    # outside context, extra validator is gone again
    Model(name="fred")


def test_unit_validation(dev_context):

    with dev_context.loader.validate_unit_models():
        # simple, one-level validation
        UnitModel(unit="Sleepy")
        with pytest.raises(ValidationError):
            UnitModel(unit="Sleepy", config={"nonexistant_param": True})

        # recursive validation
        UnitModel(
            **{
                "unit": "T3Processor",
                "config": {
                    "directives": [
                        {
                            "load": {
                                "unit": "T3SimpleDataLoader",
                                "config": {"directives": []},
                            },
                            "run": {
                                "unit": "T3UnitRunner",
                                "config": {"directives": []},
                            },
                        }
                    ]
                },
            }
        )

        with pytest.raises(ValidationError):
            UnitModel(
                **{
                    "unit": "T3Processor",
                    "config": {
                        "directives": [
                            {
                                "load": {"unit": "NotActuallyAUnit"},
                                "run": {"unit": "T3UnitRunner"},
                            }
                        ]
                    },
                }
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/T3Directive.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 04.06.2020
# Last Modified Date: 04.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Sequence
from ampel.model.UnitModel import UnitModel
from ampel.model.StrictModel import StrictModel

class T3Directive(StrictModel):
	"""
	Specification of a Tier 3 processing sequence.
	"""

	#: Provide context for run
	context: Optional[Sequence[UnitModel]]
	#: Select stocks
	select: Optional[UnitModel]
	#: Fill :class:`~ampel.core.AmpelBuffer.AmpelBuffer` for each selected stock
	load: Optional[UnitModel]
	#: Add external information to each :class:`~ampel.core.AmpelBuffer.AmpelBuffer`.
	complement: Optional[Sequence[UnitModel]]
	#: Execute unit(s) on each view (unit definition for the underlying AbsT3UnitRunner)
	run: UnitModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/compile/CompilerBase.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 23.03.2020
# Last Modified Date: 05.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Union, Set, Generator, Tuple, List
from ampel.type import ChannelId
from ampel.base import AmpelABC, defaultmethod
from ampel.model.ingest.T2IngestModel import T2IngestModel


class CompilerBase(AmpelABC):
	"""
	T2 compilers optimize the ingestion of T2 documents into the DB.
	In particular, T2 documents shared among different channels are merged with each other.
	They must be configured first using the method `add_ingest_config`.
	Then, the method `compile` can be used to optimize the creation of T2 documents.
	"""


	@defaultmethod(check_super_call=True)
	def __init__(self) -> None:
		self.t2_models: Dict[ChannelId, List[T2IngestModel]] = {}
		self.channels: Set[ChannelId] = set()


	def add_ingest_model(self, channel: ChannelId, model: T2IngestModel) -> None:

		if channel not in self.t2_models:
			self.channels.add(channel)
			self.t2_models[channel] = []

		self.t2_models[channel].append(model)


	def set_ingest_options(self,
		channel: ChannelId, model: T2IngestModel, options: Dict[str, Any]
	) -> None:
		pass


	def get_ingest_models(self,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]]
	) -> Generator[Tuple[ChannelId, T2IngestModel], None, None]:
		"""
		Get T2 configurations corresponding to the given T0 filter results
		"""

		# loop through all channels,
		for chan, res in chan_selection:

			# Skip channels unknown to/undefined by this particular ingester
			if chan not in self.channels:
				continue

			# loop through state t2 unit ids to be scheduled for this channel
			for ingest_model in self.t2_models[chan]:

				# Filters can return an int value to filter out t2s by "group id"
				if not (res is True or res in ingest_model.group): # type: ignore
					continue

				yield chan, ingest_model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/plot/SVGUtils.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 17.05.2019
# Last Modified Date: 13.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import warnings
from typing import Optional, List, Dict, Any
import zipfile, io, base64
import matplotlib as plt
from cairosvg import svg2png
from IPython.display import Image
from ampel.protocol.LoggerProtocol import LoggerProtocol
from ampel.content.SVGRecord import SVGRecord
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.PlotProperties import PlotProperties
from matplotlib.figure import Figure

# catch SyntaxWarning: "is not" with a literal. Did you mean "!="?
with warnings.catch_warnings():
	warnings.filterwarnings(
		action="ignore",
		category=SyntaxWarning,
		module=r"svgutils\.transform"
	)
	import svgutils as su

class SVGUtils:


	@staticmethod
	def mplfig_to_svg_dict(
		mpl_fig, file_name: str, title: Optional[str] = None, tags: Optional[List[str]] = None,
		compress: int = 1, width: Optional[int] = None, height: Optional[int] = None,
		close: bool = True, fig_include_title: Optional[bool] = False, logger: Optional[LoggerProtocol] = None
	) -> SVGRecord:
		"""
		:param mpl_fig: matplotlib figure
		:param tags: list of plot tags
		:param compress:
			0: no compression, 'svg' value will be a string
			1: compress svg, 'svg' value will be compressed bytes (usage: store plots into db)
			2: compress svg and include uncompressed string into key 'sgv_str'
			(useful for saving plots into db and additionaly to disk for offline analysis)
		:param width: figure width, for example 10 inches
		:param height: figure height, for example 10 inches
		:returns: svg dict instance
		"""

		if logger:
			logger.info("Saving plot %s" % file_name)

		imgdata = io.StringIO()

		if width is not None and height is not None:
			mpl_fig.set_size_inches(width, height)

		if title and fig_include_title:
			mpl_fig.suptitle(title)

		mpl_fig.savefig(
			imgdata, format='svg', bbox_inches='tight'
		)

		if close:
			plt.pyplot.close(mpl_fig)

		ret: SVGRecord = {'name': file_name}

		if tags:
			ret['tag'] = tags

		if title:
			ret['title'] = title

		if compress == 0:
			ret['compressed'] = False
			ret['svg'] = imgdata.getvalue()
			return ret

		outbio = io.BytesIO()

		zf = zipfile.ZipFile(
			outbio, "a", zipfile.ZIP_DEFLATED, False
		)

		zf.writestr(
			file_name,
			imgdata.getvalue().encode('utf8')
		)

		zf.close()

		ret['compressed'] = True
		ret['svg'] = outbio.getvalue()

		if compress == 2:
			ret['svg_str'] = imgdata.getvalue()

		return ret


	@classmethod
	def mplfig_to_svg_dict1(
		cls, mpl_fig: Figure, props: PlotProperties, extra: Optional[Dict[str, Any]] = None,
		close: bool = True, logger: Optional[LoggerProtocol] = None
	) -> SVGRecord:
		"""
		:param extra: required if file_name of title in PlotProperties use a format string ("such_%s_this")
		"""

		svg_doc = cls.mplfig_to_svg_dict(
			mpl_fig,
			file_name = props.get_file_name(extra=extra),
			title = props.get_title(extra=extra),
			fig_include_title = props.fig_include_title,
			width = props.width,
			height = props.height,
			tags = props.tags,
			compress = props.get_compress(),
			logger = logger,
			close = close
		)

		if props.disk_save:
			file_name = props.get_file_name(extra=extra)
			if logger and isinstance(logger, AmpelLogger) and logger.verbose > 1:
				logger.debug("Saving %s/%s" % (props.disk_save, file_name))
			with open("%s/%s" % (props.disk_save, file_name), "w") as f:
				f.write(
					svg_doc.pop("svg_str") # type: ignore
					if props.get_compress() == 2
					else svg_doc['svg']
				)

		return svg_doc


	@staticmethod
	def decompress_svg_dict(svg_dict: SVGRecord) -> SVGRecord:
		"""
		Modifies input dict by potentionaly decompressing compressed 'svg' value
		"""

		if not isinstance(svg_dict, dict):
			raise ValueError("Parameter svg_dict must be an instance of dict")

		if svg_dict.get('compressed'):
			bio = io.BytesIO()
			bio.write(svg_dict['svg']) # type: ignore
			zf = zipfile.ZipFile(bio)
			file_name = zf.namelist()[0]
			svg_dict['svg'] = str(zf.read(file_name), "utf8")
			del svg_dict['compressed']

		return svg_dict


	@staticmethod
	def rescale(svg: str, scale: float = 1.0) -> str:

		# Get SVGFigure from file
		original = su.transform.fromstring(svg)

		# Original size is represetnted as string (examle: '600px'); convert to float
		original_width = float(original.width.split('.')[0])
		original_height = float(original.height.split('.')[0])

		scaled = su.transform.SVGFigure(
			original_width * scale,
			original_height * scale
		)

		# Get the root element
		svg_st = original.getroot()

		# Scale the root element
		svg_st.scale_xy(scale, scale)

		# Add scaled svg element to figure
		scaled.append(svg_st)

		return str(scaled.to_str(), "utf-8")
		#return scaled.to_str()


	@staticmethod
	def to_png(content: str, dpi: int = 96) -> str:

		return '<img src="data:image/png;base64,' + str(
			base64.b64encode(
				Image(
					svg2png(
						bytestring=content,
						dpi=dpi
					),
				).data
			),
			"utf-8"
		) + '">'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/compile/PointT2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.12.2017
# Last Modified Date: 01.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Dict, Set, Optional, Union, Sequence, Any, Literal, Tuple
from ampel.content.DataPoint import DataPoint
from ampel.type import ChannelId, DataPointId
from ampel.model.ingest.T2IngestModel import T2IngestModel
from ampel.abstract.ingest.AbsPointT2Compiler import AbsPointT2Compiler


class PointT2Compiler(AbsPointT2Compiler):
	"""
	Helper class capabable of generating a nested dict that is used as basis to create T2 documents.
	The generated structure is optimized: multiple T2 documents associated with the same datapoint
	accross different channels are merged into one single T2 document
	that references all corrsponding channels.
	"""

	def __init__(self) -> None:
		super().__init__()
		self.slices: Dict[Tuple[ChannelId, str, Optional[int]], List[slice]] = {}


	def set_ingest_options(self,
		channel: ChannelId, model: T2IngestModel, options: Dict[str, Any]
	) -> None:

		s = self.get_slice(options.get('eligible'))
		k = (channel, model.unit_id, model.config)

		if k in self.slices:
			# Avoid duplicated slices from bad config
			for el in self.slices[k]:
				if s.__reduce__()[1] == el.__reduce__()[1]:
					return
			self.slices[k].append(s)
		else:
			self.slices[k] = [s]


	def get_slice(self,
		arg: Optional[Union[Literal['first', 'last'], Tuple[int, int, int]]]
	) -> slice:
		"""
		:raises: ValueError if parameter 'arg' is invalid
		"""
		if arg is None:
			return slice(None)
		elif arg == "all":
			return slice(None)
		elif arg == "first":
			return slice(1)
		elif arg == "last":
			return slice(-1, -2, -1)
		elif isinstance(arg, tuple) and len(arg) == 3:
			return slice(*arg)
		else:
			raise ValueError(
				f"Unsupported value provided as slice parameter : {arg}"
			)


	def compile(self,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
		datapoints: Sequence[DataPoint]
	) -> Dict[Tuple[str, Optional[int], DataPointId], Set[ChannelId]]:
		"""
		TLDR: This function computes and returns a dict structure helpful for creating T2 docs.
		This computation is required since:
		* A given alert can be accepted by one filter and be rejected by the other
		* T0 filters can return a customized set of T2 units to be run (group id)
		----------------------------------------------------------------------------------

		:param chan_selection: example: {"CHAN_SN": True, "CHAN_GRB": 1, "CHAN_BH": None}
		CHANNEL_SN accepted the alert and requests all associated T2s to be created
		CHANNEL_GRB accepted the alert and requests only T2s with group ID 1 to be created
		CHANNEL_BH rejected the alert

		Dict key element 1: unit id
		Dict key element 2: hashed dict value of the T2's config dict (123)
		Dict key element 3: datapoint id
		Dict values: set of channel ids
		"""

		t2s_eff: Dict[Tuple[str, Optional[int], DataPointId], Set[ChannelId]] = {}
		datapoints = list(reversed(datapoints))

		for chan, ingest_model in self.get_ingest_models(chan_selection):

			t2_id = ingest_model.unit_id
			config = ingest_model.config

			for s in self.slices[(chan, t2_id, config)]:
				for dp in datapoints[s]:
					k = (t2_id, config, dp['_id'])
					if k not in t2s_eff:
						t2s_eff[k] = {chan}
					else:
						t2s_eff[k].add(chan)

		return t2s_eff



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/UnitLoader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.10.2019
# Last Modified Date: 17.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from contextlib import contextmanager
from functools import partial
from importlib import import_module
from typing import (
	Dict, Iterator, Type, Any, Union, Optional, TypeVar, Sequence,
	Literal, List, overload, get_args, get_origin, cast
)

from pydantic.main import ModelMetaclass, create_model
from ampel.type import check_class
from ampel.util.collections import ampel_iter
from ampel.util.freeze import recursive_unfreeze
from ampel.util.mappings import merge_dicts
from ampel.util.type_analysis import get_subtype
from ampel.view.ReadOnlyDict import ReadOnlyDict
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.base.DataUnit import DataUnit
from ampel.core.AmpelContext import AmpelContext
from ampel.core.AdminUnit import AdminUnit
from ampel.model.StrictModel import StrictModel
from ampel.model.UnitModel import UnitModel
from ampel.model.Secret import Secret
from ampel.model.t3.AliasableModel import AliasableModel
from ampel.config.AmpelConfig import AmpelConfig
from ampel.log.AmpelLogger import AmpelLogger
from ampel.abstract.AbsSecretProvider import AbsSecretProvider

T = TypeVar('T', bound=AmpelBaseModel)
BT = TypeVar('BT', bound=DataUnit)
PT = TypeVar('PT', bound=AdminUnit)


class UnitLoader:

	def __init__(self,
		config: AmpelConfig,
		tier: Optional[Literal[0, 1, 2, 3]] = None,
		secrets: Optional[AbsSecretProvider] = None,
		) -> None:
		"""
		For optimization purposes, try to set the parameter tier.
		For example, a T2 controller should spawn an UnitLoader
		using UnitLoader(ampel_config, 2).

		:raises: ValueError in case bad arguments are provided
		"""

		if not isinstance(config, AmpelConfig):
			raise ValueError(
				f"First parameter must be an instance of "
				f"AmpelConfig (provided: {type(config)})"
			)

		self.ampel_config = config
		self.unit_defs: List[Dict] = [
			config._config['unit']['controller'],
			config._config['unit']['base'],
			config._config['unit']['admin'],
			config._config["unit"]["core"],
			config._config["unit"]["aux"],
		]

		self.aliases: List[Dict] = [
			config._config['alias'][f"t{el}"] for el in (0, 3, 1, 2)
		]

		self.secrets = secrets


	def resolve_aliases(self, value):
		"""
		Recursively resolve aliases from config
		"""
		if isinstance(value, str):
			for adict in self.aliases:
				if value in adict:
					return self.resolve_aliases(adict[value])
			return value
		elif isinstance(value, list):
			return [self.resolve_aliases(v) for v in value]
		elif isinstance(value, dict):
			return {k: self.resolve_aliases(v) for k, v in value.items()}
		else:
			return value

	@classmethod
	def resolve_secrets(
		cls,
		secrets: Optional[AbsSecretProvider],
		unit: Union[Type[AmpelBaseModel], ModelMetaclass],
		annotations: Dict[str, Any],
		defaults: Dict[str, Any],
		init_config: Dict[str, Any]
	) -> Dict[str, Any]:
		"""
		Recursively walk annotations, resolving any Secret fields that are found
		"""
		for field, typ in annotations.items():
			# Secret, or Union containing Secret
			if secret_field := get_subtype(Secret, typ):
				if field in init_config:
					value = init_config[field]
				elif field in defaults:
					value = defaults[field]
				else:
					raise KeyError(f"{unit.__qualname__}.{field} needs a value")

				# skip if optional or preconfigured
				if value is None and get_subtype(type(None), typ) or isinstance(value, Secret):
					continue
				elif not secrets:
					raise RuntimeError(f"{unit.__qualname__}.{field} needs a secret provider")
				elif not (isinstance(value, dict) and "key" in value):
					raise ValueError(f"{unit.__qualname__}.{field}" + " should be configured with a dict of the form {\"key\": \"secret-name\"}")
				target_type = getattr(secret_field, '__args__', [str])[0]
				init_config[field] = secrets.get(value["key"], target_type)

			# other model, possibly containing Secret fields
			elif type(typ) is ModelMetaclass:
				value = {}
				if field in init_config:
					value = init_config[field]
				elif field in defaults:
					value = defaults[field]
				init_config[field] = cls.resolve_secrets(secrets, typ, typ.__annotations__, typ.__field_defaults__, value)

			# Union, possibly containing other models
			elif get_origin(typ) is Union:
				value = {}
				if field in init_config:
					value = init_config[field]
				elif field in defaults:
					value = defaults[field]
				# skip if optional
				if value is None and type(None) in get_args(typ):
					continue
				for subtyp in get_args(typ):
					if type(subtyp) is ModelMetaclass:
						# configure for first model that satisfied by the initial value
						if isinstance(value, subtyp):
							break
						elif isinstance(value, dict) and set(value.keys()).union(subtyp.__field_defaults__) == set(subtyp.__fields__.keys()):
							init_config[field] = cls.resolve_secrets(secrets, subtyp, subtyp.__annotations__, subtyp.__field_defaults__, value)
							break
		return init_config


	def get_init_config(self,
		unit: Union[str, Type[AmpelBaseModel]],
		config: Optional[Union[int, str, Dict[str, Any]]] = None,
		override: Optional[Dict[str, Any]] = None,
		kwargs: Optional[Dict[str, Any]] = None,
		resolve_secrets = True,
	) -> Dict[str, Any]:
		""" :raises: ValueError is model type is not recognized """

		ret: Optional[Dict[str, Any]] = {}

		if isinstance(config, (dict, str)):
			ret = self.resolve_aliases(config)

		elif isinstance(config, int):
			if isinstance(
				confid := self.ampel_config.get(f"confid.{config}", dict, raise_exc=True),
				ReadOnlyDict
			):
				ret = recursive_unfreeze(confid)
			else:
				ret = confid

		if ret is None and config is not None:
			raise ValueError(f"Config alias {config} not found")

		ret = merge_dicts([ret, override, kwargs])
		if resolve_secrets and ret is not None:
			if isinstance(unit, str):
				unit = self.get_class_by_name(unit)
			if isinstance(unit, ModelMetaclass):
				annotations, defaults = unit.__annotations__, unit.__field_defaults__
			elif issubclass(unit, AmpelBaseModel):
				annotations, defaults = unit._annots, unit._defaults
			else:
				raise TypeError
			ret = self.resolve_secrets(self.secrets, unit, annotations, defaults, ret)
		return ret if ret is not None else {}


	@contextmanager
	def validate_unit_models(self) -> Iterator[None]:
		"""
		Enable validation for UnitModel instances
		"""
		extra_validator = (False, partial(_validate_unit_model, unit_loader=self))
		UnitModel.__post_root_validators__.append(extra_validator)
		AliasableModel._config = self.ampel_config
		try:
			yield
		finally:
			UnitModel.__post_root_validators__.remove(extra_validator)
			AliasableModel._config = None



	def get_resources(self, unit_model: UnitModel) -> Dict[str, Any]:
		"""
		Resources are defined using the static variable 'require' in ampel units
		-> example: catsHTM.default
		"""

		resources: Dict[str, Any] = {}

		if isinstance(unit_model.unit, str):
			unit_model.unit = self.get_class_by_name(unit_model.unit)

		# Load possibly required global resources
		for k in ampel_iter(getattr(unit_model.unit, 'require', [])):

			if k is None:
				continue

			# some unit require access to the channels definition
			if k == 'channel':
				resources[k] = self.ampel_config.get('channel')
				continue

			# Global resource example: extcat
			if (resource := self.ampel_config.get(f'resource.{k}')) is None:
				raise ValueError(f"Global resource not available: {k}")

			resources[k] = resource

		return resources


	@overload
	def get_class_by_name(self, name: str, unit_type: Type[T]) -> Type[T]:
		...
	@overload
	def get_class_by_name(self, name: str, unit_type: None = ...) -> Type[AmpelBaseModel]:
		...
	def get_class_by_name(self, name: str, unit_type: Optional[Type[T]] = None) -> Union[Type[T], Type[AmpelBaseModel]]:
		"""
		Matches the parameter 'name' with the unit definitions defined in the ampel_config.
		This allows to retrieve the corresponding fully qualified name of the class and to load it.

		:param unit_type:
			- DataUnit or any sublcass of DataUnit
			- AdminUnit or any sublcass of AdminUnit
			- If None, FQN will be retrieved from the auxiliary class conf entries and returned object will have Type[Any]

		:raises: ValueError if unit cannot be found or loaded or if parent class is unrecognized
		"""

		if name in AuxUnitRegister._defs:
			return AuxUnitRegister.get_aux_class(name, sub_type=unit_type)

		# Loop through list of class definition dicts
		fqn = None
		for udefs in self.unit_defs:
			if name in udefs:
				fqn = udefs[name]['fqn']
				break

		if fqn is None:
			raise ValueError(f"Ampel unit not found: {name}")

		# Note: importlib.import_module caches internally imported modules
		return getattr(import_module(fqn), name)


	@overload
	def new(self, unit_model: UnitModel, *, unit_type: Type[T], **kwargs) -> T:
		...
	@overload
	def new(self, unit_model: UnitModel, *, unit_type: None = ..., **kwargs) -> AmpelBaseModel:
		...
	def new(self, unit_model: UnitModel, *, unit_type: Optional[Type[T]] = None, **kwargs) -> Union[T, AmpelBaseModel]:
		"""
		Instantiate new object based on provided model and kwargs.
		:param 'unit_type': performs isinstance check and raise error on mismatch. Enables mypy/other static checks.
		"""

		if not isinstance(unit_model, UnitModel):
			raise ValueError(f"Unexpected model: '{type(unit_model)}'")

		if isinstance(unit_model.unit, str):
			unit_model.unit = self.get_class_by_name(unit_model.unit, unit_type)

		if unit_type:
			check_class(unit_model.unit, unit_type)

		init_args = self.get_init_config(
			unit_model.unit,
			unit_model.config,
			unit_model.override,
			kwargs
		)

		return unit_model.unit(**init_args) # type: ignore[call-arg]


	@overload
	def new_base_unit(self, unit_model: UnitModel, logger: AmpelLogger, *, sub_type: Type[BT], **kwargs) -> BT:
		...
	@overload
	def new_base_unit(self, unit_model: UnitModel, logger: AmpelLogger, *, sub_type: None = ..., **kwargs) -> DataUnit:
		...
	def new_base_unit(self,
		unit_model: UnitModel, logger: AmpelLogger, *, sub_type: Optional[Type[BT]] = None, **kwargs
	) -> Union[BT, DataUnit]:
		"""
		Base units require logger and resource as init parameters, additionaly to the potentialy
		defined custom parameters which will be provided as a union of the model config
		and the kwargs provided to this method (the latter having prevalance)
		:raises: ValueError is the unit defined in the model is unknown
		"""

		if sub_type is None or not issubclass(get_origin(sub_type) or sub_type, DataUnit):
			sub_type = cast(Type[BT], DataUnit) # remove cast when mypy gets smarter

		return self.new(
			unit_model, unit_type=sub_type, logger=logger, resource=self.get_resources(unit_model),
			**kwargs
		)


	@overload
	def new_admin_unit(self, unit_model: UnitModel, context: AmpelContext, *, sub_type: Type[PT], **kwargs) -> PT:
		...
	@overload
	def new_admin_unit(self, unit_model: UnitModel, context: AmpelContext, *, sub_type: None = ..., **kwargs) -> AdminUnit:
		...
	def new_admin_unit(self,
		unit_model: UnitModel, context: AmpelContext, *, sub_type: Optional[Type[PT]] = None, **kwargs
	) -> Union[AdminUnit, PT]:
		"""
		Processor units require a context as init parameters, additionaly to the potentialy
		defined custom parameters which will be provided as a union of the model config
		and the kwargs provided to this method (the latter having prevalance)
		:raises: ValueError is the unit defined in the model is unknown
		"""

		if sub_type is None or not issubclass(get_origin(sub_type) or sub_type, AdminUnit):
			sub_type = cast(Type[PT], AdminUnit) # remove cast when mypy gets smarter

		return self.new(
			unit_model, unit_type=sub_type, context=context,
			**kwargs
		)

	"""
	def internal_mypy_tests_uncomment_only_in_your_editor(self,
		model: UnitModel, context: AmpelContext, logger: AmpelLogger, sub_type: Optional[Type[PT]] = None, **kwargs
	) -> None:

		# Interal: uncomment to check if mypy works adequately

		from ampel.abstract.AbsProcessorUnit import AbsProcessorUnit
		from ampel.abstract.AbsLightCurveT2Unit import AbsLightCurveT2Unit

		reveal_type(self.new(model))
		reveal_type(self.new(model, bla=12))
		reveal_type(self.new(model, unit_type = None))
		reveal_type(self.new(model, unit_type=AbsLightCurveT2Unit))
		reveal_type(self.new(model, unit_type=AbsLightCurveT2Unit, bla=12))
		reveal_type(self.new(model, unit_type=AbsProcessorUnit))
		reveal_type(self.new(model, unit_type=AbsProcessorUnit, bla=12))

		reveal_type(self.new_base_unit(model, logger))
		reveal_type(self.new_base_unit(model, logger, bla=12))
		reveal_type(self.new_base_unit(model, logger, sub_type = None))
		reveal_type(self.new_base_unit(model, logger, sub_type=AbsLightCurveT2Unit))
		reveal_type(self.new_base_unit(model, logger, sub_type = AbsLightCurveT2Unit, bla=12))

		# Next two lines *should* fail
		reveal_type(self.new_base_unit(model, logger, sub_type=AbsProcessorUnit))
		reveal_type(self.new_base_unit(model, logger, sub_type = AbsProcessorUnit, bla=12))

		reveal_type(self.new_admin_unit(model, context))
		reveal_type(self.new_admin_unit(model, context, bla=12))
		reveal_type(self.new_admin_unit(model, context, sub_type = None))
		reveal_type(self.new_admin_unit(model, context, sub_type = AbsProcessorUnit))
		reveal_type(self.new_admin_unit(model, context, sub_type = AbsProcessorUnit, bla=12))

		# Next two lines *should* fail
		reveal_type(self.new_admin_unit(model, context, sub_type = AbsLightCurveT2Unit))
		reveal_type(self.new_admin_unit(model, context, sub_type = AbsLightCurveT2Unit, bla=12))
	"""


def _validate_unit_model(cls, values: Dict[str, Any], unit_loader: UnitLoader) -> Dict[str, Any]:
	"""
	Verify that a unit configuration is valid in the context of a specific UnitLoader.
	"""
	from ampel.base.DataUnit import DataUnit
	from ampel.core.AdminUnit import AdminUnit
	from ampel.abstract.AbsProcessorUnit import AbsProcessorUnit
	from ampel.abstract.ingest.AbsIngester import AbsIngester
	from ampel.t3.run.AbsT3UnitRunner import AbsT3UnitRunner
	from ampel.t3.context.AbsT3RunContextAppender import AbsT3RunContextAppender

	unit = unit_loader.get_class_by_name(values['unit'])
	if issubclass(unit, (DataUnit, AdminUnit, AbsProcessorUnit, AbsIngester)):
		# exclude base class fields provided at runtime
		exclude = {"logger"}
		for parent in cast(
			Sequence[Type[AmpelBaseModel]],
			(
				DataUnit,
				AdminUnit,
				AbsT3UnitRunner,
				AbsT3RunContextAppender,
				AbsProcessorUnit,
				AbsIngester,
			)
		):
			if issubclass(unit, parent):
				exclude.update(set(parent._annots.keys()).difference(parent._defaults.keys()))
		fields = {
			k: (v, unit._defaults[k] if k in unit._defaults else ...)
			for k, v in unit._annots.items() if k not in exclude
		} # type: ignore
		model = create_model(
			unit.__name__, __config__ = StrictModel.__config__,
			__base__=None, __module__=None, __validators__=None,
			**fields
		)
		model.validate(
			unit_loader.get_init_config(
				values['unit'],
				values['config'],
				values['override']
			)
		)
	return values



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/plot/SVGQuery.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.06.2019
# Last Modified Date: 03.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>


class SVGQuery:

	def __init__(
		self, col="photo", plot_path='plots',
		stock_id=None, plot_tag=None, plot_tags=None
	):

		self._query = {}
		self.tag = None
		self.tags = None
		self.plot_path = plot_path
		self.col = col

		if stock_id:
			self.set_stock_id(stock_id)

		if plot_tag:
			self.set_plot_tag(plot_tag)

		if plot_tags:
			self.set_plot_tags(plot_tags)


	def get_query(self):
		return self._query


	def set_stock_id(self, stock_id):

		if isinstance(stock_id, (list, tuple)):
			self._query['stockId'] = {'$in': stock_id}
		else:
			self._query['stockId'] = stock_id


	def set_plot_tag(self, tag):
		self.tag = tag
		self._query[self.plot_path + ".tags"] = tag


	def set_plot_tags(self, tags):
		self.tags = tags
		self._query[self.plot_path + ".tags"] = {'$all': tags}


	def set_query_parameter(self, name, value, overwrite=False):
		"""
		For example:
		set_query_parameter(
			"$or", [
				{'t2Id': 'myFirstT2', 'run_config': 'default'},
				{'t2Id': 'myT2'}
			]
		)
		"""
		if name in self._query and not overwrite:
			raise ValueError(
				"Parameter %s already defined (use overwrite=True if you know what you're doing)" % name
			)

		self._query[name] = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/operator/AllOf.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.10.2018
# Last Modified Date: 13.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pydantic.generics import GenericModel
from typing import List, Generic
from ampel.type import T


class AllOf(GenericModel, Generic[T]):
	#: Select items by logical AND
	all_of: List[T]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/db/LogsMatcher.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.11.2018
# Last Modified Date: 15.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from datetime import datetime

class LogsMatcher:
	"""
	"""

	def __init__(self):
		"""
		"""
		self.match = {}


	def get_match_criteria(self):
		"""
		:returns: dict
		"""
		return self.match


	def set_flag(self, arg):
		"""
		The mongodb operator "$bitsAllSet" will be used.

		:param LogFlag arg:
		"""
		self.match['flag'] = {"$bitsAllSet": arg}
		return self


	def set_channels(self, channels, compact_logs=True):
		"""
		:type channels: str, list[str]
		"""
		self.channels = channels
		if compact_logs:
			self.match['$or'] = [
				{'channel': channels},
				{'msg.channel': channels}
			]
		else:
			self.match['channel'] = channels
		return self


	def set_stock_ids(self, stock_ids):
		"""
		:type tran_ids: int, List[int]
		"""
		self.match['stock'] = stock_ids
		return self


	def set_run_ids(self, run_ids):
		"""
		:type run_id: int, List[int]
		"""
		self.match['run'] = run_ids
		return self


	def set_alert_ids(self, arg):
		""" """
		self.match['alert'] = arg
		return self


	def set_custom(self, key, value):
		""" """
		self.match[key] = value
		return self


	def set_after(self, dt):
		"""
		Note: time operation is greater than / *equals*
		:param dt: date-time
		:type dt: either datetime object or string (ex: '2018-06-29 08:15:27')
		"""

		self._set_time_constraint(dt, '$gte')
		return self


	def set_before(self, dt):
		"""
		Note: time operation is before than / *equals*
		:param dt: date-time
		:type dt: either datetime object or string (ex: '2018-06-29 08:15:27')
		"""
		self._set_time_constraint(dt, '$lte')
		return self


	def _set_time_constraint(self, dt, op):
		"""
		Note: time operation is greater than / *equals*
		:param dt: date-time
		:type dt: either datetime object or string (ex: '2018-06-29 08:15:27.243860')
		"""

		if isinstance(dt, datetime):
			pass
		elif isinstance(dt, str):
			dt = datetime.strptime(dt, '%Y-%m-%d %H:%M:%S')
		else:
			raise ValueError()

		from bson.objectid import ObjectId
		if "_id" not in self.match:
			self.match["_id"] = {}

		self.match["_id"][op] = ObjectId.from_datetime(dt)
		return self
